@@ -11,9 +11,9 @@ use warnings;
 use base qw( Tickit::Widget::Menu::base );
 use Tickit::Style;
 
-our $VERSION = '0.03';
+our $VERSION = '0.04';
 
-use Tickit::RenderContext qw( LINE_SINGLE );
+use Tickit::RenderBuffer qw( LINE_SINGLE );
 use Tickit::Utils qw( textwidth );
 use List::Util qw( sum max );
 
@@ -89,6 +89,24 @@ sub cols
    return sum( map { textwidth $_->name } $self->items ) + 2 * ( $self->items - 1 );
 }
 
+sub reshape
+{
+   my $self = shift;
+
+   $self->{itemcols} = \my @cols;
+   $self->{itemwidths} = \my @widths;
+
+   my $items = $self->{items};
+   my $col = 0;
+   foreach my $idx ( 0 .. $#$items ) {
+      $cols[$idx] = $col;
+      $col += ( $widths[$idx] = textwidth $items->[$idx]->name );
+      $col += 2;
+   }
+
+   push @cols, $col;
+}
+
 sub pos2item
 {
    my $self = shift;
@@ -96,18 +114,15 @@ sub pos2item
 
    $line == 0 or return ();
 
-   my @items = $self->items;
-   my $idx = 0;
-   while( $col >= 0 and $idx < @items ) {
-      my $item = $items[$idx];
+   my $cols   = $self->{itemcols};
+   my $widths = $self->{itemwidths};
 
-      my $width = textwidth $item->name;
-      return ( $item, $idx, $col ) if $col < $width;
-      $col -= $width;
-      $idx++;
+   foreach my $idx ( 0 .. $#$widths ) {
+      next unless $col < $cols->[$idx+1];
+      $col -= $cols->[$idx];
 
-      return () if $col < 2;
-      $col -= 2;
+      return () if $col >= $widths->[$idx];
+      return ( $self->{items}->[$idx], $idx, $col );
    }
 
    return ();
@@ -116,52 +131,35 @@ sub pos2item
 sub render_item
 {
    my $self = shift;
-   my ( $idx ) = @_;
+   my ( $idx, $rb ) = @_;
 
-   $self->render( first_idx => $idx, last_idx => $idx, rect => $self->window->rect );
+   my $item = $self->{items}->[$idx];
+
+   $rb->goto( 0, $self->{itemcols}->[$idx] );
+
+   my $is_highlight = defined $self->{active_idx} && $idx == $self->{active_idx};
+   $rb->text( $item->name, $is_highlight ? $self->get_style_pen( "highlight" ) : undef );
+
+   $rb->erase( 2 );
 }
 
-sub render
+sub render_to_rb
 {
    my $self = shift;
-   my %args = @_;
-
-   my $win = $self->window or return;
-   $win->is_visible or return;
-   my $rect = $args{rect};
-
-   my $rc = Tickit::RenderContext->new( lines => $win->lines, cols => $win->cols );
-   $rc->clip( $rect );
-   $rc->setpen( $self->pen );
-
-   my $highlight_pen = $self->get_style_pen( "highlight" );
+   my ( $rb, $rect ) = @_;
 
    if( $rect->top == 0 ) {
-      $rc->goto( 0, 0 );
-
       my @items = $self->items;
       foreach my $idx ( 0 .. $#items ) {
-         last if defined $args{last_idx} and $idx > $args{last_idx};
-
-         my $item = $items[$idx];
-         my $name = $item->name;
-
-         $rc->skip( textwidth( $name ) + 2 ), next if defined $args{first_idx} and $idx < $args{first_idx};
-
-         my $is_highlight = defined $self->{active_idx} && $idx == $self->{active_idx};
-
-         $rc->text( $name, $is_highlight ? $highlight_pen : undef );
-         $rc->erase( 2 );
+         $self->render_item( $idx, $rb );
       }
 
-      $rc->erase_to( $rc->cols ) if !defined $args{last_idx};
+      $rb->erase_to( $rect->right );
    }
 
-   foreach my $line ( max( $rect->top, 1 ) .. $rect->bottom ) {
-      $rc->erase_at( $line, 0, $rc->cols );
+   foreach my $line ( $rect->linerange( 1, undef ) ) {
+      $rb->erase_at( $line, $rect->left, $rect->cols );
    }
-
-   $rc->flush_to_window( $win );
 }
 
 sub popup_item
@@ -171,30 +169,45 @@ sub popup_item
 
    my $items = $self->{items};
 
-   my $col = 0;
-   for ( my $i = 0; $i < $idx; $i++ ) {
-      $col += textwidth( $items->[$i]->name ) + 2;
-   }
-
+   my $col = $self->{itemcols}->[$idx];
    $items->[$idx]->popup( $self->window, 1, $col );
 }
 
 sub activated
 {
    my $self = shift;
+   $self->dismiss;
+}
 
-   undef $self->{active_idx};
+sub dismiss
+{
+   my $self = shift;
+   $self->SUPER::dismiss;
+
+   # Still have a window after ->dismiss
    $self->redraw;
 }
 
 sub on_mouse_item
 {
    my $self = shift;
-   my ( $event, $button, $line, $col, $item, $item_idx, $item_col ) = @_;
+   my ( $args, $item, $item_idx, $item_col ) = @_;
 
-   if( $event eq "press" || $event eq "drag" and
-       $button == 1 ) {
-      $self->activate_item( $item_idx );
+   # We only ever care about button 1
+   return unless $args->button == 1;
+
+   my $event = $args->type;
+   if( $event eq "press" ) {
+      # A second click on an active item deactivates
+      if( defined $self->{active_idx} and $item_idx == $self->{active_idx} ) {
+         $self->dismiss;
+      }
+      else {
+         $self->expand_item( $item_idx );
+      }
+   }
+   elsif( $event eq "drag" ) {
+      $self->expand_item( $item_idx );
    }
 
    return 1;
